Object subclass: #FitnessStatistics
	instanceVariableNames: 'population fitnesses totalFitness maxFitness minFitness averageFitness totalScaledFitness standardDeviationFitness slope intersect '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Genetic-Core'!
!FitnessStatistics commentStamp: 'DGD 11/27/2000 19:14' prior: 0!
I'm a set of Statistics related to the fitness of a Population!


!FitnessStatistics methodsFor: 'initialization' stamp: 'DGD 11/30/2000 18:28'!
initializeOn: aPopulation 
	| delta fmultiple |
	population _ aPopulation.
	fitnesses _ aPopulation individues collect: [:individue | individue fitness].
	totalFitness _ fitnesses sum.
	maxFitness _ fitnesses max.
	minFitness _ fitnesses min.
	averageFitness _ fitnesses average.
	standardDeviationFitness _ fitnesses standardDeviation.
	"Slope and the intersect for fitness scaling 
Genetic Algorithms (Goldberg) page 79"
	fmultiple _ 2.
	slope _ 1.
	intersect _ 0.
	minFitness > (fmultiple * averageFitness - maxFitness / (fmultiple - 1))
		ifTrue: [(delta _ maxFitness - averageFitness) == 0
				ifFalse: 
					[slope _ fmultiple - 1 * averageFitness / delta.
					intersect _ averageFitness * (maxFitness - (fmultiple * averageFitness)) / delta]]
		ifFalse: [(delta _ averageFitness - minFitness) == 0
				ifFalse: 
					[slope _ averageFitness / delta.
					intersect _ minFitness negated * (averageFitness / delta)]]! !


!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:35'!
averageFitness
	^averageFitness! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:59'!
fitnesses
	^fitnesses! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/27/2000 18:27'!
intersect
	^intersect! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:59'!
maxFitness
	^maxFitness! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:54'!
minFitness
	^minFitness! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/27/2000 18:27'!
slope
	^slope! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/27/2000 18:47'!
standardDeviationFitness
	^standardDeviationFitness! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:35'!
totalFitness
	^totalFitness! !

!FitnessStatistics methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:53'!
totalScaledFitness
	totalScaledFitness isNil ifTrue: [

	population individues do: [:individue | individue calculateScaledFitness: self ].
	totalScaledFitness _ (population individues detectSum: [:individue | individue scaledFitness])

].
	^ totalScaledFitness! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

FitnessStatistics class
	instanceVariableNames: ''!

!FitnessStatistics class methodsFor: 'instance creation' stamp: 'DGD 11/28/2000 18:25'!
newOn: aPopulation 
	^ self new initializeOn: aPopulation! !


Model subclass: #GeneticAlgorithm
	instanceVariableNames: 'population bestIndividue goal start '
	classVariableNames: 'GARandom '
	poolDictionaries: ''
	category: 'Genetic-Core'!
!GeneticAlgorithm commentStamp: '<historical>' prior: 0!
The algorithm!


!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 10/1/2000 22:00'!
bestIndividue
	^bestIndividue! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 11/28/2000 22:03'!
bestIndividue: anObject 
	bestIndividue _ anObject. Transcript show: '#' , self population generation printString , ' Best: ' , anObject printString;
			 cr! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 10/1/2000 21:53'!
goal
	^goal! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 10/1/2000 21:53'!
goal: anObject
	goal _ anObject! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 10/1/2000 21:53'!
population
	^population! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 11/25/2000 01:56'!
population: aPopulation 
	population _ aPopulation.
	self bestIndividue isNil
		ifTrue: [self bestIndividue: aPopulation bestIndividue]
		ifFalse: [aPopulation bestIndividue fitness > self bestIndividue fitness ifTrue: [self bestIndividue: aPopulation bestIndividue. self changed: #bestIndividue]].
	self changed: #population! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 11/28/2000 18:58'!
start
	^start! !

!GeneticAlgorithm methodsFor: 'accessing' stamp: 'DGD 11/28/2000 18:58'!
start: anObject
	start _ anObject! !


!GeneticAlgorithm methodsFor: 'running' stamp: 'DGD 11/30/2000 23:58'!
run
self changed: #bestIndividue.
self changed: #population.
	[self goal value: self population]
		whileFalse: [self population: self population next]! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

GeneticAlgorithm class
	instanceVariableNames: ''!

!GeneticAlgorithm class methodsFor: 'testing' stamp: 'DGD 12/1/2000 23:28'!
test
	| ga |
	ga _ GeneticAlgorithm population: (Population newRandom: 50 especie: OnesCountIndividue)
				goal: [:p | p individues anySatisfy: [:i | i genoma allSatisfy: [:c | c = 1]]].
	(GeneticAlgorithmMorph newOn: ga) openInWorld.
	ga run! !


!GeneticAlgorithm class methodsFor: 'instance creation' stamp: 'DGD 11/28/2000 18:54'!
population: aPopulation goal: aBlock 
	^ self new population: aPopulation;
	 goal: aBlock;
start: Time now;
	 yourself! !


!GeneticAlgorithm class methodsFor: 'accessing' stamp: 'DGD 10/9/2000 23:14'!
random
	^ GARandom! !


!GeneticAlgorithm class methodsFor: 'initialize-release' stamp: 'DGD 11/24/2000 18:52'!
initialize
	GARandom _ GeneticAlgorithmRandom new! !


Random subclass: #GeneticAlgorithmRandom
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Genetic-Core'!
!GeneticAlgorithmRandom commentStamp: 'DGD 11/27/2000 19:15' prior: 0!
I'm a Random specialized on Genetic Algorithms!


!GeneticAlgorithmRandom methodsFor: 'genetic' stamp: 'DGD 11/24/2000 19:13'!
atRandom: aSequenceableCollection 
	^ aSequenceableCollection atRandom:self! !

!GeneticAlgorithmRandom methodsFor: 'genetic' stamp: 'DGD 11/30/2000 17:28'!
crossoverProbability
	^ self next < 0.9! !

!GeneticAlgorithmRandom methodsFor: 'genetic' stamp: 'DGD 11/24/2000 18:51'!
locus:anInteger
	^ self nextInt:anInteger! !

!GeneticAlgorithmRandom methodsFor: 'genetic' stamp: 'DGD 12/1/2000 16:23'!
muteProbability
	^ self next < 0.003! !

!GeneticAlgorithmRandom methodsFor: 'genetic' stamp: 'DGD 11/24/2000 18:54'!
rouletteWheelValue: anNumber 
	^ self next * anNumber! !


Object subclass: #Individue
	instanceVariableNames: 'genoma cachedFitness scaledFitness '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Genetic-Core'!
!Individue commentStamp: '<historical>' prior: 0!
Im an Individue of a Population!


!Individue methodsFor: 'accessing' stamp: 'DGD 11/30/2000 19:56'!
cachedFitness: anObject
	cachedFitness _ anObject! !

!Individue methodsFor: 'accessing' stamp: 'DGD 11/30/2000 17:33'!
genoma
	^genoma! !

!Individue methodsFor: 'accessing' stamp: 'DGD 12/1/2000 17:29'!
genoma: anObject 

genoma = anObject ifFalse:[

	genoma _ anObject.
	self cachedFitness: nil
]! !

!Individue methodsFor: 'accessing' stamp: 'DGD 11/26/2000 22:43'!
scaledFitness
	^scaledFitness! !


!Individue methodsFor: 'genetic-fitness' stamp: 'DGD 10/3/2000 00:05'!
calculateFitness
	self subclassResponsibility! !

!Individue methodsFor: 'genetic-fitness' stamp: 'DGD 11/30/2000 18:25'!
calculateScaledFitness: aFitnessStatistics 
"
TODO: Refactor as Strategy
"
	"self linearScaling: aFitnessStatistics"
	self sigmaScaling: aFitnessStatistics! !

!Individue methodsFor: 'genetic-fitness' stamp: 'DGD 11/21/2000 19:55'!
fitness
	cachedFitness isNil
		ifTrue: [cachedFitness _ self calculateFitness].
	^ cachedFitness! !

!Individue methodsFor: 'genetic-fitness' stamp: 'DGD 11/30/2000 18:27'!
linearScaling: aFitnessStatistics 
	"Genetic Algorithms (Goldberg) page 79"
	scaledFitness _ aFitnessStatistics slope * self fitness + aFitnessStatistics intersect + (1 / 10)! !

!Individue methodsFor: 'genetic-fitness' stamp: 'DGD 11/30/2000 18:53'!
sigmaScaling: aFitnessStatistics 
	"An Introduction to Genetic Algorithms (Mitchell) page 167"
	| fitnessPremium sigma |
	sigma _ aFitnessStatistics standardDeviationFitness.
	fitnessPremium _ sigma isZero
				ifTrue: [1]
				ifFalse: [self fitness - aFitnessStatistics averageFitness / (2 * sigma)].
	scaledFitness _ fitnessPremium max: 1 / 10! !


!Individue methodsFor: 'genetic-operators' stamp: 'DGD 11/24/2000 18:25'!
crossover: anIndividue
self subclassResponsibility! !

!Individue methodsFor: 'genetic-operators' stamp: 'DGD 11/24/2000 18:07'!
mutate
	self subclassResponsibility! !


!Individue methodsFor: 'memetic-operators' stamp: 'DGD 12/1/2000 19:31'!
neighborhood
	"^ #()"
self subclassResponsibility! !

!Individue methodsFor: 'memetic-operators' stamp: 'DGD 12/1/2000 22:32'!
searchLocal
	| current bestNeighbor inLocal | 
			current _ self.
			inLocal _ false.
			
			[bestNeighbor _ current neighborhood detectMax: [:neighbor | neighbor fitness].
			(bestNeighbor isNil not and: [bestNeighbor fitness > current fitness])
				ifTrue: [current _ bestNeighbor]
				ifFalse: [inLocal _ true].
			inLocal] whileFalse.
			self == current ifFalse: [self become: current]! !

!Individue methodsFor: 'memetic-operators' stamp: 'DGD 12/1/2000 22:32'!
trySearchLocal
	self class searchLocal
		ifTrue: 
			[self searchLocal]! !


!Individue methodsFor: 'printing' stamp: 'DGD 12/1/2000 23:37'!
printOn: aStream 
	"super printOn: aStream.  
	aStream nextPutAll: ', genoma='."
	
	self genoma do: [:c | c printOn: aStream].
	aStream nextPutAll: ', f='.
	self fitness printOn: aStream! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Individue class
	instanceVariableNames: ''!

!Individue class methodsFor: 'genetic' stamp: 'DGD 12/1/2000 19:31'!
searchLocal
^ false! !


!Individue class methodsFor: 'instance creation' stamp: 'DGD 11/30/2000 17:36'!
genoma: anObject 
	^ self new genoma: anObject;
	 yourself! !

!Individue class methodsFor: 'instance creation' stamp: 'DGD 10/9/2000 23:06'!
newRandom
	self subclassResponsibility! !


Individue subclass: #ListIndividue
	instanceVariableNames: ''
	classVariableNames: 'OthersAlleles '
	poolDictionaries: ''
	category: 'Genetic-Core'!
!ListIndividue commentStamp: '<historical>' prior: 0!
I represent the Individues based on Lists.!


!ListIndividue methodsFor: 'genetic-operators' stamp: 'DGD 11/30/2000 19:01'!
crossover: anIndividue 
	| locus aNewIndividue otherNewIndividue |
	locus _ GeneticAlgorithm random crossoverProbability
				ifTrue: [(GeneticAlgorithm random locus: self genoma size)
						roundTo: self class crossoverPoint]
				ifFalse: [0].
	aNewIndividue _ self class genoma: (self genoma first: locus)
					, (anIndividue genoma allButFirst: locus).
	otherNewIndividue _ self class genoma: (anIndividue genoma first: locus)
					, (self genoma allButFirst: locus).
	^ {aNewIndividue. otherNewIndividue}! !

!ListIndividue methodsFor: 'genetic-operators' stamp: 'DGD 12/1/2000 16:06'!
mutate
	| random |
	random _ GeneticAlgorithm random.
	self genoma
		withIndexDo: [:gen :locus | random muteProbability
				ifTrue: 
					[self cachedFitness: nil.
					self genoma at: locus put: (random atRandom: (self class othersAlleles:  gen))]]! !


!ListIndividue methodsFor: 'memetic-operators' stamp: 'DGD 12/1/2000 18:57'!
neighborhood
	| neighborhood |
	neighborhood _ OrderedCollection new.
	genoma withIndexDo: [:gen :locus | (self class othersAlleles: gen)
			do: 
				[:newGen | 
				| newGenoma | 
				newGenoma _ genoma veryDeepCopy.
				newGenoma at: locus put: newGen.
				neighborhood add: (self species genoma: newGenoma)]].
	^ neighborhood! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

ListIndividue class
	instanceVariableNames: ''!

!ListIndividue class methodsFor: 'genetic' stamp: 'DGD 11/24/2000 18:12'!
allele
	self subclassResponsibility! !

!ListIndividue class methodsFor: 'genetic' stamp: 'DGD 11/29/2000 18:09'!
crossoverPoint
 ^ 1! !

!ListIndividue class methodsFor: 'genetic' stamp: 'DGD 11/24/2000 18:11'!
listSize
	self subclassResponsibility! !

!ListIndividue class methodsFor: 'genetic' stamp: 'DGD 12/1/2000 16:15'!
othersAlleles: anAllele 
	^ OthersAlleles at:anAllele ifAbsentPut:[self allele copyWithout: anAllele]! !


!ListIndividue class methodsFor: 'initialize-release' stamp: 'DGD 12/1/2000 19:21'!
initialize
OthersAlleles _ Dictionary new.! !


!ListIndividue class methodsFor: 'instance creation' stamp: 'DGD 12/1/2000 19:21'!
newRandom
	self initialize.
	^ self new genoma: (Array streamContents: [:s | self listSize timesRepeat: [s nextPut: (GeneticAlgorithm random atRandom: self allele)]]);
	 yourself! !


ListIndividue subclass: #OnesCountIndividue
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Genetic-Core'!
!OnesCountIndividue commentStamp: 'DGD 12/1/2000 23:25' prior: 0!
Im an Individue of 0 & 1 and my fitness is the count of 1!


!OnesCountIndividue methodsFor: 'genetic-fitness' stamp: 'DGD 11/30/2000 17:53'!
calculateFitness
	^ self genoma sum! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

OnesCountIndividue class
	instanceVariableNames: ''!

!OnesCountIndividue class methodsFor: 'genetic' stamp: 'DGD 12/1/2000 19:32'!
allele
	^ #(0 1 )! !

!OnesCountIndividue class methodsFor: 'genetic' stamp: 'DGD 12/1/2000 19:38'!
listSize
	^ 130! !


!OnesCountIndividue class methodsFor: 'testing' stamp: 'DGD 12/1/2000 23:28'!
test
	| i1 i2 |
	Transcript clear.
	i1 _ OnesCountIndividue genoma: #(1 1 1 1 1 1 1 1 ).
	i2 _ OnesCountIndividue genoma: #(0 0 0 0 0 0 0 0 ).
	20 timesRepeat: [Transcript show: (i1 crossover: i2);
		 cr]! !

!OnesCountIndividue class methodsFor: 'testing' stamp: 'DGD 12/1/2000 23:28'!
test2
	Time millisecondsToRun: [(OnesCountIndividue genoma: #(0 0 0 0 0 0 0 0 )) searchLocal]! !

!OnesCountIndividue class methodsFor: 'testing' stamp: 'DGD 12/1/2000 23:28'!
test3
	(OnesCountIndividue genoma: #(0 0 0 0 0 0 0 0 )) neighborhood! !


Object subclass: #Population
	instanceVariableNames: 'individues generation fitnessStatistics millisecondsToRun '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Genetic-Core'!
!Population commentStamp: '<historical>' prior: 0!
The Population of Individues!


!Population methodsFor: 'accessing' stamp: 'DGD 11/28/2000 18:26'!
fitnessStatistics
	fitnessStatistics isNil ifTrue: [fitnessStatistics _ FitnessStatistics newOn: self].
	^ fitnessStatistics! !

!Population methodsFor: 'accessing' stamp: 'DGD 10/2/2000 22:29'!
generation
	^generation! !

!Population methodsFor: 'accessing' stamp: 'DGD 10/2/2000 22:29'!
generation: anObject
	generation _ anObject! !

!Population methodsFor: 'accessing' stamp: 'DGD 10/1/2000 21:14'!
individues
	^individues! !

!Population methodsFor: 'accessing' stamp: 'DGD 10/1/2000 17:02'!
individues: anObject
	individues _ anObject! !

!Population methodsFor: 'accessing' stamp: 'DGD 11/29/2000 13:23'!
millisecondsToRun
	^millisecondsToRun! !

!Population methodsFor: 'accessing' stamp: 'DGD 11/29/2000 13:10'!
millisecondsToRun: anObject
	millisecondsToRun _ anObject! !


!Population methodsFor: 'genetic' stamp: 'DGD 11/24/2000 16:33'!
bestIndividue
	^ individues first! !

!Population methodsFor: 'genetic' stamp: 'DGD 12/1/2000 22:32'!
next
	| nextGeneration millis nextIndividues parent1 parent2 |
	nextGeneration _ self species new.
	nextGeneration generation: self generation + 1.
	millis _ Time
				millisecondsToRun: 
					[nextIndividues _ Bag new.
					self individues size / 2
						timesRepeat: 
							[parent1 _ self selectIndividue.
							parent2 _ self selectIndividue.
							nextIndividues addAll: (parent1 crossover: parent2)].
					nextIndividues do: 
						[:individue | 
						individue mutate.
						individue trySearchLocal].
					nextGeneration individues addAll: nextIndividues].
	nextGeneration millisecondsToRun: millis.
	^ nextGeneration! !

!Population methodsFor: 'genetic' stamp: 'DGD 11/26/2000 22:49'!
selectIndividue
	self individues inject: (GeneticAlgorithm random rouletteWheelValue: self fitnessStatistics totalScaledFitness)
		into: 
			[:rand :individue | 
			rand <= individue scaledFitness ifTrue: [^ individue].
			rand - individue scaledFitness].
	^ self individues anyOne! !


!Population methodsFor: 'initialization' stamp: 'DGD 11/28/2000 21:55'!
initialize
	self individues: (SortedCollection sortBlock: [:x :y | x fitness >= y fitness])! !


!Population methodsFor: 'printing' stamp: 'DGD 11/26/2000 22:34'!
printOn: aStream 
	aStream print: self class;
	 nextPutAll: '#';
	 print: self generation;
	 nextPutAll: ', fitness=';
	 print: self fitnessStatistics totalFitness;
	 nextPutAll: ', individues=';
	 print: self individues size;
	 nextPutAll: ', average fitness=';
	 print: (self fitness / self individues size) asFloat! !

"-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- "!

Population class
	instanceVariableNames: ''!

!Population class methodsFor: 'instance creation' stamp: 'DGD 10/9/2000 23:46'!
new

	^ super new initialize! !

!Population class methodsFor: 'instance creation' stamp: 'DGD 11/30/2000 18:30'!
newRandom: anInteger especie: anIndividueClass 
	| population millis |
	millis _ Time
				millisecondsToRun: 
					[population _ self new.
					(anInteger roundTo: 2 )timesRepeat: [population individues add: anIndividueClass newRandom].
					population generation: 1].
	population millisecondsToRun: millis.
	^ population! !

GeneticAlgorithm initialize!
ListIndividue initialize!